"""This module contains all the objects that can be used in the game window.
Each object is a subclass of Sprite, which is a subclass of pygame.sprite.Sprite.
The objects are: Box, Circle, Line, Text, and Group.
Each object has a corresponding new_* function that can be used to create the object.
For example, play.new_box() creates a new Box object.
"""

from .box import Box
from .circle import Circle
from .line import Line
from .sprite import Sprite
from .text import Text
from .image import Image



"""This module contains functions and decorators for handling key presses."""

import pygame

from ..utils.async_helpers import _make_async

pygame.key.set_repeat(200, 16)

_pressed_keys_subscriptions = {}
_release_keys_subscriptions = {}

_pressed_keys = []

_keys_released_this_frame = []
_keys_to_skip = (pygame.K_MODE,)
pygame.event.set_allowed(
    [
        pygame.QUIT,
        pygame.KEYDOWN,
        pygame.KEYUP,
        pygame.MOUSEBUTTONDOWN,
        pygame.MOUSEBUTTONUP,
        pygame.MOUSEMOTION,
    ]
)


def when_any_key(func, released=False):
    """Run a function when any key is pressed or released."""
    async_callback = _make_async(func)

    async def wrapper(*args, **kwargs):
        wrapper.is_running = True
        await async_callback(*args, **kwargs)
        wrapper.is_running = False

    wrapper.keys = None
    wrapper.is_running = False
    if released:
        if "any" not in _release_keys_subscriptions:
            _release_keys_subscriptions["any"] = []
        _release_keys_subscriptions["any"].append(wrapper)
    else:
        if "any" not in _pressed_keys_subscriptions:
            _pressed_keys_subscriptions["any"] = []
        _pressed_keys_subscriptions["any"].append(wrapper)
    return wrapper


def when_key(*keys, released=False):
    """Run a function when a key is pressed or released."""
    for key in keys:
        if not isinstance(key, str) and not (isinstance(key, list) and (not released)):
            print(key)
            raise ValueError("Key must be a string or a list of strings.")
        if isinstance(key, list):
            for sub_key in key:
                if not isinstance(sub_key, str):
                    raise ValueError("Key must be a string or a list of strings.")

    def decorator(func):
        async_callback = _make_async(func)

        async def wrapper(*args, **kwargs):
            wrapper.is_running = True
            await async_callback(*args, **kwargs)
            wrapper.is_running = False

        wrapper.is_running = False

        for key in keys:
            if isinstance(key, list):
                key = hash(frozenset(key))
            if released:
                if key not in _release_keys_subscriptions:
                    _release_keys_subscriptions[key] = []
                _release_keys_subscriptions[key].append(wrapper)
            else:
                if key not in _pressed_keys_subscriptions:
                    _pressed_keys_subscriptions[key] = []
                _pressed_keys_subscriptions[key].append(wrapper)
        return wrapper

    return decorator


def key_num_to_name(pygame_key_event):
    """Convert a pygame key event to a human-readable string."""
    return pygame.key.name(pygame_key_event.key)



"""Generate random numbers, colors, and positions."""

import random as _random

from ..io import screen
from ..utils import _Position


def random_number(lowest=0, highest=100):
    """Return a random number between `lowest` and `highest`.
    :param lowest: The lowest number that can be returned.
    :param highest: The highest number that can be returned.
    :return: A random number between `lowest` and `highest`.
    """
    # if user supplies whole numbers, return whole numbers
    if isinstance(lowest, int) and isinstance(highest, int):
        return _random.randint(lowest, highest)
    # if user supplied any floats, return decimals
    return round(_random.uniform(lowest, highest), 2)


def random_color():
    """Return a random color.
    :return: A random color.
    """
    return random_number(0, 255), random_number(0, 255), random_number(0, 255)


def random_position(
    x_min=screen.left, x_max=screen.right, y_min=screen.bottom, y_max=screen.top
):
    """
    Returns a random position on the screen. A position has an `x` and `y` e.g.:
        position = play.random_position()
        sprite.x = position.x
        sprite.y = position.y

    or equivalently:
        sprite.go_to(play.random_position())
    """
    return _Position(
        random_number(x_min, x_max),
        random_number(y_min, y_max),
    )



"""This module contains code to help with async functions."""

import asyncio as _asyncio
import warnings as _warnings


def _raise_on_await_warning(func):
    """
    If someone doesn't put 'await' before functions that require 'await'
    like play.timer() or play.animate(), raise an exception.
    :param func: A function that may or may not be async
    """

    async def raise_on_warning(*args, **kwargs):
        with _warnings.catch_warnings(record=True) as warnings:
            await func(*args, **kwargs)
            for warning in warnings:
                str_message = warning.message.args[
                    0
                ]  # e.g. "coroutine 'timer' was never awaited"
                if "was never awaited" in str_message:
                    unawaited_function_name = str_message.split("'")[1]

                    raise AssertionError(
                        f"""Looks like you forgot to put "await" before play.{unawaited_function_name}"""
                        + """on line {warning.lineno} of file {warning.filename}."""
                        + """To fix this, just add the word 'await' before play.{unawaited_function_name} on line {
                        warning.lineno} of file {warning.filename} in the function {func.__name__}."""  # pylint:
                        # disable=line-too-long
                        # pylint: enable=line-too-long
                    )
                print(warning.message)

    return raise_on_warning


def _make_async(func):
    """
    Turn a non-async function into an async function.
    Used mainly in decorators like @repeat_forever.
    :param func: A function that may or may not be async.
    """
    if _asyncio.iscoroutinefunction(func):
        # if it's already async just return it
        return _raise_on_await_warning(func)

    @_raise_on_await_warning
    async def async_func(*args, **kwargs):
        return func(*args, **kwargs)

    return async_func



"""All the events that can be triggered in the game."""

import logging as _logging

import pygame  # pylint: disable=import-error
from play.core import (
    game_loop as _game_loop,
    _repeat_forever_callbacks,
    _when_program_starts_callbacks,
)
from ..io.keypress import (
    when_key as _when_key,
    when_any_key as _when_any_key,
)
from ..io.mouse import mouse
from ..utils.async_helpers import _make_async
from ..loop import loop as _loop


# @decorator
def when_program_starts(func):
    """
    Call code right when the program starts.

    Used like this:

    @play.when_program_starts
    def do():
        print('the program just started!')
    :param func: The function to call when the program starts.
    :return: The decorator function.
    """
    async_callback = _make_async(func)

    async def wrapper(*args, **kwargs):
        return await async_callback(*args, **kwargs)

    _when_program_starts_callbacks.append(wrapper)
    return func


def repeat(number_of_times):
    """
    Repeat a set of commands a certain number of times.

    Equivalent to `range(1, number_of_times+1)`.

    Used like this:

    @play.repeat_forever
    async def do():
        for count in play.repeat(10):
            print(count)
    :param number_of_times: The number of times to repeat the commands.
    :return: A range object that can be iterated over.
    """
    return range(1, number_of_times + 1)


def start_program():
    """
    Calling this function starts your program running.

    play.start_program() should almost certainly go at the very end of your program.
    """
    for func in _when_program_starts_callbacks:
        _loop.create_task(func())

    _loop.call_soon(_game_loop)
    try:
        _loop.run_forever()
    finally:
        _logging.getLogger("asyncio").setLevel(_logging.CRITICAL)
        pygame.quit()  # pylint: disable=no-member


# @decorator
def repeat_forever(func):
    """
    Calls the given function repeatedly in the game loop.

    Example:

        text = play.new_text(words='hi there!', x=0, y=0, font='Arial.ttf', font_size=20, color='black')

        @play.repeat_forever
        async def do():
            text.turn(degrees=15)
    :param func: The function to call repeatedly.
    :return: The decorator function.
    """
    async_callback = _make_async(func)

    async def repeat_wrapper():
        repeat_wrapper.is_running = True
        await async_callback()
        repeat_wrapper.is_running = False

    repeat_wrapper.is_running = False
    _repeat_forever_callbacks.append(repeat_wrapper)
    return func


# @decorator
def when_sprite_clicked(*sprites):
    """A decorator that runs a function when a sprite is clicked.
    :param sprites: The sprites to run the function on.
    :return: The function to run.
    """

    def wrapper(func):
        for sprite in sprites:
            sprite.when_clicked(func, call_with_sprite=True)
        return func

    return wrapper


# @decorator
def when_any_key_pressed(func):
    """
    Calls the given function when any key is pressed.
    """
    if not callable(func):
        raise ValueError(
            """@play.when_any_key_pressed doesn't use a list of keys. Try just this instead:

@play.when_any_key_pressed
async def do(key):
    print("This key was pressed!", key)
"""
        )
    return _when_any_key(func, released=False)


# @decorator
def when_key_pressed(*keys):
    """
    Calls the given function when any of the specified keys are pressed.
    """
    return _when_key(*keys, released=False)


# @decorator
def when_any_key_released(func):
    """
    Calls the given function when any key is released.
    """
    if not callable(func):
        raise ValueError(
            """@play.when_any_key_released doesn't use a list of keys. Try just this instead:

@play.when_any_key_released
async def do(key):
    print("This key was released!", key)
"""
        )
    return _when_any_key(func, released=True)


# @decorator
def when_key_released(*keys):
    """
    Calls the given function when any of the specified keys are released.
    """
    return _when_key(*keys, released=True)


# @decorator
def when_mouse_clicked(func):
    """
    Calls the given function when the mouse is clicked.
    """
    return mouse.when_clicked(func)


# @decorator
def when_click_released(func):
    """
    Calls the given function when the mouse click is released.
    """
    return mouse.when_click_released(func)



"""
This module contains the API for the game.
"""

from pygame import init  # pylint: disable=no-name-in-module,import-error

from .generators import *
from .events import *
from .utils import *
from .random import *

init()



"""This file contains all the IO-related code for the Python Play library."""

from sys import platform

import pygame
import pymunk as _pymunk
from pygame._sdl2.video import Window  # pylint: disable=no-name-in-module
from pygame.locals import *

from ..globals import _walls
from ..physics import physics_space

PYGAME_DISPLAY = None


class Screen:
    def __init__(self, width=800, height=600):
        global PYGAME_DISPLAY
        self._width = width
        self._height = height
        PYGAME_DISPLAY = pygame.display.set_mode(
            (width, height), pygame.DOUBLEBUF  # pylint: disable=no-member
        )  # pylint: disable=no-member
        pygame.display.set_caption("Python Play")
        self._fullscreen = False

    @property
    def width(self):
        """Get the width of the screen.
        :return: The width of the screen."""
        return self._width

    @width.setter
    def width(self, _width):
        """Set the width of the screen.
        :param _width: The new width of the screen."""
        global PYGAME_DISPLAY
        self._width = _width

        remove_walls()
        create_walls()

        if self._fullscreen:
            self.enable_fullscreen()
        else:
            PYGAME_DISPLAY = pygame.display.set_mode((self._width, self._height))

    @property
    def height(self):
        """Get the height of the screen.
        :return: The height of the screen."""
        return self._height

    @height.setter
    def height(self, _height):
        """Set the height of the screen.
        :param _height: The new height of the screen."""
        global PYGAME_DISPLAY
        self._height = _height

        remove_walls()
        create_walls()

        if self._fullscreen:
            self.enable_fullscreen()
        else:
            PYGAME_DISPLAY = pygame.display.set_mode((self._width, self._height))

    @property
    def top(self):
        """Get the top side of the screen.
        :return: The top side of the screen."""
        return self.height / 2

    @property
    def bottom(self):
        """Get the bottom side of the screen.
        :return: The bottom side of the screen."""
        return self.height / -2

    @property
    def left(self):
        """Get the left side of the screen.
        :return: The left side of the screen."""
        return self.width / -2

    @property
    def right(self):
        """Get the right side of the screen.
        :return: The right side of the screen."""
        return self.width / 2

    @property
    def size(self):
        """Get the size of the screen.
        :return: The size of the screen."""
        return self.width, self.height

    def enable_fullscreen(self):
        """Enable fullscreen mode."""
        global PYGAME_DISPLAY
        if self._fullscreen:
            return
        self._fullscreen = True
        window = Window.from_display_module()
        full_screen_size = (
            pygame.display.Info().current_w,
            pygame.display.Info().current_h,
        )
        if platform != "linux":
            pygame.display.toggle_fullscreen()  # works for entering and exiting fullscreen, except in linux
            window.position = (
                full_screen_size[i] / 2 - window.size[i] / 2 for i in range(2)
            )  # reset X and Y position of the window to original instead of top left
        else:
            PYGAME_DISPLAY = pygame.display.set_mode(
                (self.width, self.height),
                SCALED + NOFRAME + FULLSCREEN,  # pylint: disable=undefined-variable
                32,  # pylint: disable=undefined-variable
            )  # all flags are necessary

    def disable_fullscreen(self):
        """Disable fullscreen mode."""
        global PYGAME_DISPLAY
        if not self._fullscreen:
            return
        self._fullscreen = False
        pygame.display.quit()
        pygame.display.init()
        PYGAME_DISPLAY = pygame.display.set_mode((self.width, self.height))


screen = Screen()


def _create_wall(a, b):
    segment = _pymunk.Segment(physics_space.static_body, a, b, 0.0)
    segment.elasticity = 1.0
    segment.friction = 0.1
    physics_space.add(segment)
    return segment


def create_walls():
    """Create walls around the screen."""
    _walls.append(
        _create_wall([screen.left, screen.top], [screen.right, screen.top])
    )  # top
    _walls.append(
        _create_wall([screen.left, screen.bottom], [screen.right, screen.bottom])
    )  # bottom
    _walls.append(
        _create_wall([screen.left, screen.bottom], [screen.left, screen.top])
    )  # left
    _walls.append(
        _create_wall([screen.right, screen.bottom], [screen.right, screen.top])
    )  # right


def remove_walls():
    """Remove the walls from the physics space."""
    physics_space.remove(_walls)
    _walls.clear()


def remove_wall(index):
    """Remove a wall from the physics space.
    :param index: The index of the wall to remove. 0: top, 1: bottom, 2: left, 3: right.
    """
    physics_space.remove(_walls[index])
    _walls.pop(index)


create_walls()


def convert_pos(x, y):
    """
    Convert from the Play coordinate system to the Pygame coordinate system.
    :param x: The x-coordinate in the Play coordinate system.
    :param y: The y-coordinate in the Play coordinate system.
    """
    x1 = screen.width / 2 + x
    y1 = screen.height / 2 - y
    return x1, y1


def pos_convert(x, y):
    """
    Convert from the Pygame coordinate system to the Play coordinate system.
    :param x: The x-coordinate in the Pygame coordinate system.
    :param y: The y-coordinate in the Pygame coordinate system.
    """
    x1 = x - screen.width / 2
    y1 = screen.height / 2 - y
    return x1, y1



"""A simple database system for storing data in a JSON file."""

import json
import os

FILENAME = "database.json"

JSON_DATA = None

if not os.path.exists(FILENAME):
    with open(FILENAME, "w", encoding="utf-8") as write_f:
        write_f.write("{}")

with open(FILENAME, "r", encoding="utf-8") as read_file:
    data = read_file.read()
    JSON_DATA = json.loads(data)


def get_data(key):
    """Get a value from the database.
    :param key: The key to get the value from, for which the : is a delimiter for nested values.
    """
    keys = key.split(":")
    value = JSON_DATA
    for k in keys:
        value = value[k]
    return value


def set_data(key, value):
    """Set a value in the database.
    :param key: The key to set the value to, for which the : is a delimiter for nested values.
    :param value: The value to set.
    """
    keys = key.split(":")
    target = JSON_DATA
    for k in keys[:-1]:
        if k not in target:
            raise KeyError(f"Key {k} not found in {target}")
        target = target[k]
    target[keys[-1]] = value

    with open(FILENAME, "w", encoding="utf-8") as write_file:
        write_file.write(json.dumps(JSON_DATA))



"""This module defines the Circle class, which represents a circle in the game."""

import pygame
from .sprite import Sprite
from ..io import convert_pos
from ..utils import color_name_to_rgb as _color_name_to_rgb


class Circle(Sprite):
    def __init__(  # pylint: disable=too-many-arguments
        self,
        color="black",
        x=0,
        y=0,
        radius=100,
        border_color="light blue",
        border_width=0,
        transparency=100,
        size=100,
        angle=0,
    ):
        super().__init__()
        self._x = x
        self._y = y
        self._color = color
        self._radius = radius
        self._border_color = border_color
        self._border_width = border_width

        self._transparency = transparency
        self._size = size
        self._angle = angle
        self._is_clicked = False
        self._is_hidden = False
        self.physics = None

        self._when_clicked_callbacks = []

        self.rect = pygame.Rect(0, 0, 0, 0)
        self.update()

    def clone(self):
        """Create a copy of the circle.
        :return: A copy of the circle."""
        return self.__class__(
            color=self.color,
            radius=self.radius,
            border_color=self.border_color,
            border_width=self.border_width,
            **self._common_properties()
        )

    def update(self):
        """Update the circle's position, size, angle, and transparency."""
        if self._should_recompute:
            self._image = pygame.Surface(
                (self._radius * 2, self._radius * 2), pygame.SRCALPHA
            )
            pygame.draw.circle(
                self._image,
                _color_name_to_rgb(self._color),
                (self._radius, self._radius),
                self._radius,
            )
            self.rect = self._image.get_rect()
            pos = convert_pos(self.x, self.y)
            self.rect.x = pos[0] - self._radius
            self.rect.y = pos[1] - self._radius
            super().update()

    ##### color #####
    @property
    def color(self):
        """The color of the circle.
        :return: The color of the circle."""
        return self._color

    @color.setter
    def color(self, _color):
        """Set the color of the circle.
        :param _color: The color of the circle."""
        self._color = _color

    ##### radius #####
    @property
    def radius(self):
        """The radius of the circle.
        :return: The radius of the circle."""
        return self._radius

    @radius.setter
    def radius(self, _radius):
        """Set the radius of the circle.
        :param _radius: The radius of the circle."""
        self._radius = _radius
        if self.physics:
            self.physics._pymunk_shape.unsafe_set_radius(self._radius)

    ##### border_color #####
    @property
    def border_color(self):
        """The color of the circle's border.
        :return: The color of the circle's border."""
        return self._border_color

    @border_color.setter
    def border_color(self, _border_color):
        """Set the color of the circle's border.
        :param _border_color: The color of the circle's border."""
        self._border_color = _border_color

    ##### border_width #####
    @property
    def border_width(self):
        """The width of the circle's border.
        :return: The width of the circle's border."""
        return self._border_width

    @border_width.setter
    def border_width(self, _border_width):
        """Set the width of the circle's border.
        :param _border_width: The width of the circle's border."""
        self._border_width = _border_width



"""This module contains the Mouse class and the mouse object."""

import math as _math

from ..utils.async_helpers import _make_async
from ..objects.sprite import point_touching_sprite


class _Mouse:
    def __init__(self):
        self.x = 0
        self.y = 0
        self._is_clicked = False
        self._when_clicked_callbacks = []
        self._when_click_released_callbacks = []

    @property
    def is_clicked(self):
        """Return whether the mouse is clicked.
        :return: True if the mouse is clicked, False otherwise."""
        return self._is_clicked

    def is_touching(self, other):
        """Check if the mouse is touching a sprite.
        :param other: The sprite to check.
        :return: True if the mouse is touching the sprite, False otherwise."""
        return point_touching_sprite(self, other)

    # @decorator
    def when_clicked(self, func):
        """Run a function when the mouse is clicked.
        :param func: The function to run."""
        async_callback = _make_async(func)

        async def wrapper():
            await async_callback()

        self._when_clicked_callbacks.append(wrapper)
        return wrapper

    # @decorator
    def when_click_released(self, func):
        """Run a function when the mouse click is released.
        :param func: The function to run."""
        async_callback = _make_async(func)

        async def wrapper():
            await async_callback()

        self._when_click_released_callbacks.append(wrapper)
        return wrapper

    def distance_to(self, x=None, y=None):
        """Get the distance from the mouse to a point.
        :param x: The x-coordinate of the point.
        :param y: The y-coordinate of the point.
        :return: The distance from the mouse to the point."""
        assert x is not None

        try:
            # x can either be a number or a sprite. If it's a sprite:
            x = x.x
            y = x.y
        except AttributeError:
            pass

        dx = self.x - x
        dy = self.y - y

        return _math.sqrt(dx**2 + dy**2)


mouse = _Mouse()



"""Core game loop and event handling functions."""

import math as _math

import pygame  # pylint: disable=import-error

from ..globals import backdrop, FRAME_RATE, sprites_group
from ..io import screen, PYGAME_DISPLAY
from ..io.keypress import (
    key_num_to_name as _pygame_key_to_name,
    _keys_released_this_frame,
    _keys_to_skip,
    _pressed_keys,
    _pressed_keys_subscriptions,
    _release_keys_subscriptions,
)  # don't pollute user-facing namespace with library internals
from ..io.mouse import mouse
from ..objects.line import Line
from ..objects.sprite import point_touching_sprite
from ..physics import simulate_physics
from ..utils import color_name_to_rgb as _color_name_to_rgb
from ..loop import loop as _loop

_repeat_forever_callbacks = []
_when_program_starts_callbacks = []
_clock = pygame.time.Clock()

click_happened_this_frame = False  # pylint: disable=invalid-name
click_release_happened_this_frame = False  # pylint: disable=invalid-name


def _handle_pygame_events():
    """Handle pygame events in the game loop."""
    global click_happened_this_frame
    global click_release_happened_this_frame

    for event in pygame.event.get():
        if event.type == pygame.QUIT or (  # pylint: disable=no-member
            event.type == pygame.KEYDOWN  # pylint: disable=no-member
            and event.key == pygame.K_q  # pylint: disable=no-member
            and (
                pygame.key.get_mods() & pygame.KMOD_META  # pylint: disable=no-member
                or pygame.key.get_mods() & pygame.KMOD_CTRL  # pylint: disable=no-member
            )
        ):
            # quitting by clicking window's close button or pressing ctrl+q / command+q
            _loop.stop()
            return False
        if event.type == pygame.MOUSEBUTTONDOWN:  # pylint: disable=no-member
            click_happened_this_frame = True
            mouse._is_clicked = True
        if event.type == pygame.MOUSEBUTTONUP:  # pylint: disable=no-member
            click_release_happened_this_frame = True
            mouse._is_clicked = False
        if event.type == pygame.MOUSEMOTION:  # pylint: disable=no-member
            mouse.x, mouse.y = (event.pos[0] - screen.width / 2.0), (
                screen.height / 2.0 - event.pos[1]
            )
        if event.type == pygame.KEYDOWN:  # pylint: disable=no-member
            if event.key not in _keys_to_skip:
                name = _pygame_key_to_name(event)
                if name not in _pressed_keys:
                    _pressed_keys.append(name)
        if event.type == pygame.KEYUP:  # pylint: disable=no-member
            name = _pygame_key_to_name(event)
            if not (event.key in _keys_to_skip) and name in _pressed_keys:
                _keys_released_this_frame.append(name)
                _pressed_keys.remove(name)
    return True


# pylint: disable=too-many-branches
def _handle_keyboard():
    """Handle keyboard events in the game loop."""
    ############################################################
    # @when_any_key_pressed and @when_key_pressed callbacks
    ############################################################
    if _pressed_keys:
        for key in _pressed_keys:
            if key in _pressed_keys_subscriptions:
                for callback in _pressed_keys_subscriptions[key]:
                    if not callback.is_running:
                        _loop.create_task(callback(key))
            if "any" in _pressed_keys_subscriptions:
                for callback in _pressed_keys_subscriptions["any"]:
                    if not callback.is_running:
                        _loop.create_task(callback(key))
        keys_hash = hash(frozenset(_pressed_keys))
        if keys_hash in _pressed_keys_subscriptions:
            for callback in _pressed_keys_subscriptions[keys_hash]:
                if not callback.is_running:
                    _loop.create_task(callback(_pressed_keys))

    ############################################################
    # @when_any_key_released and @when_key_released callbacks
    ############################################################
    for key in _keys_released_this_frame:
        if key in _release_keys_subscriptions:
            for callback in _release_keys_subscriptions[key]:
                if not callback.is_running:
                    _loop.create_task(callback(key))
        if "any" in _release_keys_subscriptions:
            for callback in _release_keys_subscriptions["any"]:
                if not callback.is_running:
                    _loop.create_task(callback(key))


def _handle_mouse_loop():
    """Handle mouse events in the game loop."""
    ####################################
    # @mouse.when_clicked callbacks
    ####################################
    if mouse._when_clicked_callbacks:
        for callback in mouse._when_clicked_callbacks:
            _loop.create_task(callback())

    ########################################
    # @mouse.when_click_released callbacks
    ########################################
    if mouse._when_click_released_callbacks:
        for callback in mouse._when_click_released_callbacks:
            _loop.create_task(callback())


def _update_sprites():
    for sprite in sprites_group.sprites():
        sprite._is_clicked = False
        if sprite.is_hidden:
            continue

        ######################################################
        # update sprites with results of physics simulation
        ######################################################
        if sprite.physics and sprite.physics.can_move:

            body = sprite.physics._pymunk_body
            angle = _math.degrees(body.angle)
            if isinstance(sprite, Line):
                sprite._x = body.position.x - (sprite.length / 2) * _math.cos(angle)
                sprite._y = body.position.y - (sprite.length / 2) * _math.sin(angle)
                sprite._x1 = body.position.x + (sprite.length / 2) * _math.cos(angle)
                sprite._y1 = body.position.y + (sprite.length / 2) * _math.sin(angle)
                # sprite._length, sprite._angle = sprite._calc_length_angle()
            else:
                if (
                    str(body.position.x) != "nan"
                ):  # this condition can happen when changing sprite.physics.can_move
                    sprite._x = body.position.x
                if str(body.position.y) != "nan":
                    sprite._y = body.position.y

            sprite.angle = (
                angle  # needs to be .angle, not ._angle so surface gets recalculated
            )
            sprite.physics._x_speed, sprite.physics._y_speed = body.velocity

        #################################
        # @sprite.when_clicked events
        #################################
        if mouse.is_clicked:

            if (
                point_touching_sprite((mouse.x, mouse.y), sprite)
                and click_happened_this_frame
            ):
                # only run sprite clicks on the frame the mouse was clicked
                sprite._is_clicked = True
                for callback in sprite._when_clicked_callbacks:
                    if not callback.is_running:
                        _loop.create_task(callback())

        #################################
        # @sprite.when_touching events
        #################################
        if sprite._active_callbacks:
            for cb in sprite._active_callbacks:
                _loop.create_task(cb())

    sprites_group.update()
    sprites_group.draw(PYGAME_DISPLAY)


# pylint: disable=too-many-branches, too-many-statements
def game_loop():
    """The main game loop."""
    _keys_released_this_frame.clear()
    global click_happened_this_frame, click_release_happened_this_frame
    click_happened_this_frame = False
    click_release_happened_this_frame = False

    _clock.tick(FRAME_RATE)

    if not _handle_pygame_events():
        return False

    _handle_keyboard()

    if click_happened_this_frame or click_release_happened_this_frame:
        _handle_mouse_loop()

    #############################
    # @repeat_forever callbacks
    #############################
    for callback in _repeat_forever_callbacks:
        if not callback.is_running:
            _loop.create_task(callback())

    #############################
    # physics simulation
    #############################
    _loop.call_soon(simulate_physics)

    PYGAME_DISPLAY.fill(_color_name_to_rgb(backdrop))

    _update_sprites()

    pygame.display.flip()
    _loop.call_soon(game_loop)
    return True



"""This module contains the base sprite class for all objects in the game."""

import math as _math
import warnings as _warnings
import pymunk as _pymunk
import pygame

from ..loop import loop as _loop
from ..globals import sprites_group
from ..physics import physics_space, Physics as _Physics
from ..utils import _clamp
from ..io import screen
from ..utils.async_helpers import _make_async


def _sprite_touching_sprite(a, b):
    """Check if two sprites are touching.
    :param a: The first sprite to check if it's touching the other sprite.
    :param b: The second sprite to check if it's touching the other sprite.
    :return: Whether the two sprites are touching."""
    return a.rect.colliderect(b.rect)


def point_touching_sprite(point, sprite):
    """Check if a point is touching a sprite.
    :param point: The point to check if it's touching the sprite.
    :param sprite: The sprite to check if it's touching the point.
    :return: Whether the point is touching the sprite."""
    return sprite.rect.collidepoint(point)


_should_ignore_update = ["_should_recompute", "rect", "_image", "image"]


class Sprite(
    pygame.sprite.Sprite
):  # pylint: disable=attribute-defined-outside-init, too-many-public-methods
    def __init__(self, image=None):
        self._size = None
        self._x = None
        self._y = None
        self._angle = None
        self._transparency = None

        self._when_touching_callbacks = []
        self._dependent_sprites = []
        self._active_callbacks = []

        self._image = image
        self.physics = None
        self._is_clicked = False
        self._is_hidden = False
        self._should_recompute = True

        self._when_clicked_callbacks = []

        self.rect = None

        super().__init__()
        sprites_group.add(self)

    def __setattr__(self, name, value):
        # ignore if it's in the ignored list or if the variable doesn't change
        if name not in _should_ignore_update and getattr(self, name, value) != value:
            self._should_recompute = True
            for sprite in self._dependent_sprites:
                sprite._should_recompute = True
        super().__setattr__(name, value)

    def update(self):
        """Update the sprite."""
        if self._should_recompute and self._when_touching_callbacks:
            # check if we are touching any other sprites
            for callback, b in self._when_touching_callbacks:
                if self.is_touching(b):
                    if callback not in self._active_callbacks:
                        self._active_callbacks.append(callback)
                else:
                    if callback in self._active_callbacks:
                        self._active_callbacks.remove(callback)
            self._should_recompute = False

    @property
    def is_clicked(self):
        """Get whether the sprite is clicked.
        :return: Whether the sprite is clicked."""
        return self._is_clicked

    def move(self, steps=3):
        """Move the sprite.
        :param steps: The number of steps to move the sprite."""
        angle = _math.radians(self.angle)
        self.x += steps * _math.cos(angle)
        self.y += steps * _math.sin(angle)

    def turn(self, degrees=10):
        """Turn the sprite.
        :param degrees: The number of degrees to turn the sprite."""
        self.angle += degrees

    @property
    def x(self):
        """Get the x-coordinate of the sprite.
        :return: The x-coordinate of the sprite."""
        return self._x

    @x.setter
    def x(self, _x):
        """Set the x-coordinate of the sprite.
        :param _x: The x-coordinate of the sprite."""
        prev_x = self._x
        self._x = _x
        if self.physics:
            self.physics._pymunk_body.position = self._x, self._y
            if prev_x != _x:
                # setting velocity makes the simulation more realistic usually
                self.physics._pymunk_body.velocity = (
                    _x - prev_x,
                    self.physics._pymunk_body.velocity.y,
                )
            if self.physics._pymunk_body.body_type == _pymunk.Body.STATIC:
                physics_space.reindex_static()

    @property
    def y(self):
        """Get the y-coordinate of the sprite.
        :return: The y-coordinate of the sprite."""
        return self._y

    @y.setter
    def y(self, _y):
        """Set the y-coordinate of the sprite.
        :param _y: The y-coordinate of the sprite."""
        prev_y = self._y
        self._y = _y
        if self.physics:
            self.physics._pymunk_body.position = self._x, self._y
            if prev_y != _y:
                # setting velocity makes the simulation more realistic usually
                self.physics._pymunk_body.velocity = (
                    self.physics._pymunk_body.velocity.x,
                    _y - prev_y,
                )
            if self.physics._pymunk_body.body_type == _pymunk.Body.STATIC:
                physics_space.reindex_static()

    @property
    def transparency(self):
        """Get the transparency of the sprite.
        :return: The transparency of the sprite."""
        return self._transparency

    @transparency.setter
    def transparency(self, alpha):
        """Set the transparency of the sprite.
        :param alpha: The transparency of the sprite."""
        if not isinstance(alpha, float) and not isinstance(alpha, int):
            raise ValueError(
                f"""Looks like you're trying to set {self}'s transparency to '{alpha}', which isn't a number.
Try looking in your code for where you're setting transparency for {self} and change it a number.
"""
            )
        if alpha > 100 or alpha < 0:
            _warnings.warn(
                f"""The transparency setting for {self} is being set to {alpha} and it should be between 0 and 100.
You might want to look in your code where you're setting transparency and make sure it's between 0 and 100.  """,
                Warning,
            )

        self._transparency = _clamp(alpha, 0, 100)

    @property
    def image(self):
        """Get the image of the sprite.
        :return: The image of the sprite."""
        return self._image

    @image.setter
    def image(self, image_filename):
        """Set the image of the sprite.
        :param image_filename: The filename of the image to set."""
        self._image = image_filename

    @property
    def angle(self):
        """Get the angle of the sprite.
        :return: The angle of the sprite."""
        return self._angle

    @angle.setter
    def angle(self, _angle):
        """Set the angle of the sprite.
        :param _angle: The angle of the sprite."""
        self._angle = _angle

        if self.physics:
            self.physics._pymunk_body.angle = _math.radians(_angle)

    @property
    def size(self):
        """Get the size of the sprite.
        :return: The size of the sprite."""
        return self._size

    @size.setter
    def size(self, percent):
        """Set the size of the sprite.
        :param percent: The size of the sprite as a percentage."""
        self._size = percent
        if self.physics:
            self.physics._remove()
            self.physics._make_pymunk()

    def hide(self):
        """Hide the sprite."""
        self._is_hidden = True
        if self.physics:
            self.physics.pause()

    def show(self):
        """Show the sprite."""
        self._is_hidden = False
        if self.physics:
            self.physics.unpause()

    @property
    def is_hidden(self):
        """Get whether the sprite is hidden.
        :return: Whether the sprite is hidden."""
        return self._is_hidden

    @is_hidden.setter
    def is_hidden(self, hide):
        """Set whether the sprite is hidden.
        :param hide: Whether the sprite is hidden."""
        self._is_hidden = hide

    @property
    def is_shown(self):
        """Get whether the sprite is shown.
        :return: Whether the sprite is shown."""
        return not self._is_hidden

    @is_shown.setter
    def is_shown(self, show):
        """Set whether the sprite is shown.
        :param show: Whether the sprite is shown."""
        self._is_hidden = not show

    def is_touching(self, sprite_or_point):
        """Check if the sprite is touching another sprite or a point.
        :param sprite_or_point: The sprite or point to check if it's touching.
        :return: Whether the sprite is touching the other sprite or point."""
        if isinstance(sprite_or_point, Sprite):
            return _sprite_touching_sprite(self, sprite_or_point)
        return point_touching_sprite(sprite_or_point, self)

    def point_towards(self, x, y=None):
        """Point the sprite towards a point or another sprite.
        :param x: The x-coordinate of the point.
        :param y: The y-coordinate of the point."""
        try:
            x, y = x.x, x.y
        except AttributeError:
            pass
        self.angle = _math.degrees(_math.atan2(y - self.y, x - self.x))

    def go_to(self, x=None, y=None):
        """
        Example:

            # text will follow around the mouse
            text = play.new_text('yay')

            @play.repeat_forever
            async def do():
                text.go_to(play.mouse)
        """
        assert not x is None

        try:
            # users can call e.g. sprite.go_to(play.mouse), so x will be an object with x and y
            self.x = x.x
            self.y = x.y
        except AttributeError:
            self.x = x
            self.y = y

    def distance_to(self, x, y=None):
        """Calculate the distance to a point or sprite.
        :param x: The x-coordinate of the point.
        :param y: The y-coordinate of the point.
        :return: The distance to the point or sprite."""
        assert not x is None

        try:
            # x can either be a number or a sprite. If it's a sprite:
            x1 = x.x
            y1 = x.y
        except AttributeError:
            x1 = x
            y1 = y

        dx = self.x - x1
        dy = self.y - y1

        return _math.sqrt(dx**2 + dy**2)

    def remove(self):
        """Remove the sprite from the screen."""
        if self.physics:
            self.physics._remove()
        sprites_group.remove(self)

    @property
    def width(self):
        """Get the width of the sprite.
        :return: The width of the sprite."""
        return self.rect.width

    @property
    def height(self):
        """Get the height of the sprite.
        :return: The height of the sprite."""
        return self.rect.height

    @property
    def right(self):
        """Get the right of the sprite.
        :return: The right of the sprite."""
        return self.x + self.width / 2

    @right.setter
    def right(self, x):
        """Set the right of the sprite to a x-coordinate.
        :param x: The x-coordinate to set the right of the sprite to."""
        self.x = x - self.width / 2

    @property
    def left(self):
        """Get the left of the sprite.
        :return: The left of the sprite."""
        return self.x - self.width / 2

    @left.setter
    def left(self, x):
        """Set the left of the sprite to a x-coordinate.
        :param x: The x-coordinate to set the left of the sprite to."""
        self.x = x + self.width / 2

    @property
    def top(self):
        """Get the top of the sprite.
        :return: The top of the sprite."""
        return self.y + self.height / 2

    @top.setter
    def top(self, y):
        """Set the top of the sprite to a y-coordinate.
        :param y: The y-coordinate to set the top of the sprite to."""
        self.y = y - self.height / 2

    @property
    def bottom(self):
        """Get the bottom of the sprite.
        :return: The bottom of the sprite."""
        return self.y - self.height / 2

    @bottom.setter
    def bottom(self, y):
        """Set the bottom of the sprite to a y-coordinate.
        :param y: The y-coordinate to set the bottom of the sprite to."""
        self.y = y + self.height / 2

    def _pygame_x(self):
        return self.x + (screen.width / 2.0) - (self.rect.width / 2.0)

    def _pygame_y(self):
        return (screen.height / 2.0) - self.y - (self.rect.height / 2.0)

    # @decorator
    def when_clicked(self, callback, call_with_sprite=False):
        """Run a function when the sprite is clicked.
        :param callback: The function to run.
        :param call_with_sprite: Whether to call the function with the sprite as an argument.
        """
        async_callback = _make_async(callback)

        async def wrapper():
            wrapper.is_running = True
            if call_with_sprite:
                await async_callback(self)
            else:
                await async_callback()
            wrapper.is_running = False

        wrapper.is_running = False
        self._when_clicked_callbacks.append(wrapper)
        return wrapper

    def when_touching(self, *sprites):
        """Run a function when the sprite is touching another sprite.
        :param sprites: The sprites to check if they're touching.
        """

        def decorator(func):
            async_callback = _make_async(func)

            async def wrapper(*args, **kwargs):
                wrapper.is_running = True
                await async_callback(*args, **kwargs)
                wrapper.is_running = False

            wrapper.is_running = False

            for sprite in sprites:
                print(sprite)
                sprite._dependent_sprites.append(self)
                self._when_touching_callbacks.append((wrapper, sprite))
            self._when_touching_callbacks = frozenset(self._when_touching_callbacks)
            print(self._when_touching_callbacks)
            return wrapper

        return decorator

    def _common_properties(self):
        # used with inheritance to clone
        return {
            "x": self.x,
            "y": self.y,
            "size": self.size,
            "transparency": self.transparency,
            "angle": self.angle,
        }

    def clone(self):
        """Clone the sprite.
        :return: The cloned sprite."""
        # TODO: make work with physics
        return self.__class__(image=self.image)

    def start_physics(  # pylint: disable=too-many-arguments
        self,
        can_move=True,
        stable=False,
        x_speed=0,
        y_speed=0,
        obeys_gravity=True,
        bounciness=1.0,
        mass=10,
        friction=0.1,
    ):
        """Start the physics simulation for this sprite.
        :param can_move: Whether the object can move.
        :param stable: Whether the object is stable.
        :param x_speed: The x-speed of the object.
        :param y_speed: The y-speed of the object.
        :param obeys_gravity: Whether the object obeys gravity.
        :param bounciness: The bounciness of the object.
        :param mass: The mass of the object.
        :param friction: The friction of the object.
        """
        if not self.physics:
            self.physics = _Physics(
                self,
                can_move,
                stable,
                x_speed,
                y_speed,
                obeys_gravity,
                bounciness,
                mass,
                friction,
            )

    def stop_physics(self):
        """Stop the physics simulation for this sprite."""
        self.physics._remove()
        self.physics = None



"""This module contains the Line class, which is a subclass of Sprite. It is used to create lines in the game window."""

import math as _math

import pygame
from .sprite import Sprite
from ..io import convert_pos, screen
from ..utils import color_name_to_rgb as _color_name_to_rgb


class Line(Sprite):
    def __init__(  # pylint: disable=too-many-arguments
        self,
        color="black",
        x=0,
        y=0,
        length=None,
        angle=None,
        thickness=1,
        x1=None,
        y1=None,
        transparency=100,
        size=100,
    ):
        super().__init__()
        self._x = x
        self._y = y
        self._color = color
        self._thickness = thickness

        # can set either (length, angle) or (x1,y1), otherwise a default is used
        if length is not None and angle is not None:
            self._length = length
            self._angle = angle
            self._x1, self._y1 = self._calc_endpoint()
        elif x1 is not None and y1 is not None:
            self._x1 = x1
            self._y1 = y1
            self._length, self._angle = self._calc_length_angle()
        else:
            # default values
            self._length = length or 100
            self._angle = angle or 0
            self._x1, self._y1 = self._calc_endpoint()

        self._transparency = transparency
        self._size = size

        self.rect = pygame.Rect(0, 0, 0, 0)
        self.update()

    def update(self):
        """Update the line's position and appearance."""
        if self._should_recompute:
            pos_begin = convert_pos(self.x, self.y)
            pos_end = convert_pos(self.x1, self.y1)

            self._image = pygame.Surface((screen.width, screen.height), pygame.SRCALPHA)
            pygame.draw.line(
                self._image,
                _color_name_to_rgb(self._color),
                pos_begin,
                pos_end,
                self._thickness,
            )
            self.rect = self._image.get_rect()
            super().update()

    def clone(self):
        """Return a clone of the line.
        :return: A clone of the line."""
        return self.__class__(
            color=self.color,
            length=self.length,
            thickness=self.thickness,
            **self._common_properties()
        )

    ##### color #####
    @property
    def color(self):
        """Return the color of the line.
        :return: The color of the line."""
        return self._color

    @color.setter
    def color(self, _color):
        """Set the color of the line.
        :param _color: The new color of the line."""
        self._color = _color

    ##### thickness #####
    @property
    def thickness(self):
        """Return the thickness of the line.
        :return: The thickness of the line."""
        return self._thickness

    @thickness.setter
    def thickness(self, _thickness):
        """Set the thickness of the line.
        :param _thickness: The new thickness of the line."""
        self._thickness = _thickness

    def _calc_endpoint(self):
        radians = _math.radians(self._angle)
        return (
            self._length * _math.cos(radians) + self.x,
            self._length * _math.sin(radians) + self.y,
        )

    ##### length #####
    @property
    def length(self):
        """Return the length of the line.
        :return: The length of the line
        :rtype: float"""
        return self._length

    @length.setter
    def length(self, _length):
        """Set the length of the line.
        :param _length: The new length of the line."""
        self._length = _length
        self._x1, self._y1 = self._calc_endpoint()

    ##### angle #####
    @property
    def angle(self):
        """Return the angle of the line.
        :return: The angle of the line."""
        return self._angle

    @angle.setter
    def angle(self, _angle):
        """Set the angle of the line.
        :param _angle: The new angle of the line."""
        self._angle = _angle
        self._x1, self._y1 = self._calc_endpoint()
        if self.physics:
            self.physics._pymunk_body.angle = _math.radians(_angle)

    def _calc_length_angle(self):
        dx = self.x1 - self.x
        dy = self.y1 - self.y

        # TODO: this doesn't work at all
        return _math.sqrt(dx**2 + dy**2), _math.degrees(_math.atan2(dy, dx))

    ##### x1 #####
    @property
    def x1(self):
        """Return the x-coordinate of the line's endpoint.
        :return: The x-coordinate of the line's endpoint."""
        return self._x1

    @x1.setter
    def x1(self, _x1):
        """Set the x-coordinate of the line's endpoint.
        :param _x1: The new x-coordinate of the line's endpoint."""
        self._x1 = _x1
        self._length, self._angle = self._calc_length_angle()

    ##### y1 #####
    @property
    def y1(self):
        """Return the y-coordinate of the line's endpoint.
        :return: The y-coordinate of the line's endpoint."""
        return self._y1

    @y1.setter
    def y1(self, _y1):
        """Set the y-coordinate of the line's endpoint.
        :param _y1: The new y-coordinate of the line's endpoint."""
        self._y1 = _y1
        self._length, self._angle = self._calc_length_angle()



"""This module contains the Image class, which is a subclass of the Sprite class."""

import pygame

from .sprite import Sprite
from ..io import convert_pos


class Image(Sprite):
    def __init__(
        self, image, x=0, y=0, angle=0, size=100, transparency=100
    ):  # pylint: disable=too-many-arguments
        super().__init__()
        if isinstance(image, str):
            image = pygame.image.load(image)
        self._image = image
        self._x = x
        self._y = y
        self._angle = angle
        self._size = size
        self._transparency = transparency
        self.rect = self._image.get_rect()
        self.update()

    def update(self):
        """Update the image's position, size, angle, and transparency."""
        if self._should_recompute:
            self._image = pygame.transform.scale(self._image, (self.width, self.height))
            self._image = pygame.transform.rotate(self._image, self.angle)
            self._image.set_alpha(self.transparency)
            self.rect = self._image.get_rect()
            pos = convert_pos(self.x, self.y)
            self.rect.center = pos
            super().update()



"""Game functions and utilities."""

import asyncio as _asyncio
import logging as _logging

import pygame  # pylint: disable=import-error

from .events import _when_program_starts_callbacks, _game_loop
from ..loop import loop as _loop
from ..utils import color_name_to_rgb as _color_name_to_rgb
from ..io.keypress import _pressed_keys
from ..globals import backdrop as __backdrop

_backdrop = __backdrop  # Work around for the global variable not being imported


def start_program():
    """
    Calling this function starts your program running.

    play.start_program() should almost certainly go at the very end of your program.
    """
    for func in _when_program_starts_callbacks:
        _loop.create_task(func())

    _loop.call_soon(_game_loop)
    try:
        _loop.run_forever()
    finally:
        _logging.getLogger("asyncio").setLevel(_logging.CRITICAL)
        pygame.quit()  # pylint: disable=no-member


def stop_program():
    """
    Calling this function stops your program running.

    play.stop_program() should almost certainly go at the very end of your program.
    """
    _loop.stop()
    pygame.quit()  # pylint: disable=no-member


async def animate():
    """
    Wait for the next frame to be drawn.
    """
    await _asyncio.sleep(0)


def set_backdrop(color_or_image_name):
    """Set the backdrop color or image for the game.
    :param color_or_image_name: The color or image to set as the backdrop.
    """
    global _backdrop
    _color_name_to_rgb(color_or_image_name)

    _backdrop = color_or_image_name


async def timer(seconds=1.0):
    """Wait a number of seconds. Used with the await keyword like this:

    @play.repeat_forever
    async def do():
        await play.timer(seconds=2)
        print('hi')
    :param seconds: The number of seconds to wait.
    :return: True after the number of seconds has passed.
    """
    await _asyncio.sleep(seconds)
    return True


def key_is_pressed(*keys):
    """
    Returns True if any of the given keys are pressed.

    Example:

        @play.repeat_forever
        async def do():
            if play.key_is_pressed('up', 'w'):
                print('up or w pressed')
    """

    for key in keys:
        if key in _pressed_keys.values():
            return True
    return False



"""This module contains the Box class, which represents a box in the game."""

import pygame

from .sprite import Sprite
from ..io import convert_pos
from ..utils import color_name_to_rgb as _color_name_to_rgb


class Box(Sprite):
    def __init__(  # pylint: disable=too-many-arguments
        self,
        color="black",
        x=0,
        y=0,
        width=100,
        height=200,
        border_color="light blue",
        border_width=0,
        transparency=100,
        size=100,
        angle=0,
    ):
        super().__init__(self)
        self._color = color
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._border_color = border_color
        self._border_width = border_width
        self._transparency = transparency
        self._size = size
        self._angle = angle
        self.rect = pygame.Rect(0, 0, 0, 0)
        self.update()

    def update(self):
        """Update the box's position, size, angle, and transparency."""
        if self._should_recompute:
            self.image = pygame.Surface((self._width, self._height), pygame.SRCALPHA)
            self.image.fill(_color_name_to_rgb(self._color))
            self.image.set_alpha(self._transparency)
            self.rect = self.image.get_rect()
            pos = convert_pos(self.x, self.y)
            self.rect.x = pos[0] - self._width // 2
            self.rect.y = pos[1] - self._height // 2
            super().update()

    ##### width #####
    @property
    def width(self):
        """The width of the box.
        :return: The width of the box."""
        return self._width

    @width.setter
    def width(self, _width):
        """Set the width of the box.
        :param _width: The new width of the box."""
        self._width = _width

    ##### height #####
    @property
    def height(self):
        """The height of the box.
        :return: The height of the box."""
        return self._height

    @height.setter
    def height(self, _height):
        """Set the height of the box.
        :param _height: The new height of the box."""
        self._height = _height

    ##### color #####
    @property
    def color(self):
        """The color of the box.
        :return: The color of the box."""
        return self._color

    @color.setter
    def color(self, _color):
        """Set the color of the box.
        :param _color: The new color of the box."""
        self._color = _color

    ##### border_color #####
    @property
    def border_color(self):
        """The color of the box's border.
        :return: The color of the box's border."""
        return self._border_color

    @border_color.setter
    def border_color(self, _border_color):
        """Set the color of the box's border.
        :param _border_color: The new color of the box's border."""
        self._border_color = _border_color

    ##### border_width #####
    @property
    def border_width(self):
        """The width of the box's border.
        :return: The width of the box's border."""
        return self._border_width

    @border_width.setter
    def border_width(self, _border_width):
        """Set the width of the box's border.
        :param _border_width: The new width of the box's border."""
        self._border_width = _border_width

    def clone(self):
        """Create a copy of the box.
        :return: A copy of the box."""
        return self.__class__(
            color=self.color,
            width=self.width,
            height=self.height,
            border_color=self.border_color,
            border_width=self.border_width,
            **self._common_properties()
        )



"""This module contains the Text class, which is used to create text objects in the game."""

import pygame
from .sprite import Sprite
from ..io import convert_pos
from ..utils import color_name_to_rgb as _color_name_to_rgb


class Text(Sprite):
    def __init__(  # pylint: disable=too-many-arguments
        self,
        words="hi :)",
        x=0,
        y=0,
        font=None,
        font_size=50,
        color="black",
        angle=0,
        transparency=100,
        size=100,
    ):
        super().__init__()
        self._font = font
        self._font_size = font_size
        self._pygame_font = pygame.font.Font(font, font_size)
        self._words = words
        self._color = color

        self._x = x
        self._y = y

        self._size = size
        self._angle = angle
        self.transparency = transparency

        self._is_clicked = False
        self._is_hidden = False
        self.physics = None

        self._when_clicked_callbacks = []

        self.rect = pygame.Rect(0, 0, 0, 0)
        self.update()

    def update(self):
        """Update the text object."""
        if self._should_recompute:
            pos = convert_pos(self.x, self.y)
            self._image = self._pygame_font.render(
                self._words, True, _color_name_to_rgb(self._color)
            )
            self.rect = self.image.get_rect()
            self.rect.topleft = (
                pos[0] - self.rect.width // 2,
                pos[1] - self.rect.height // 2,
            )
            super().update()

    def clone(self):
        return self.__class__(
            words=self.words,
            font=self.font,
            font_size=self.font_size,
            color=self.color,
            **self._common_properties(),
        )

    @property
    def words(self):
        """Get the words of the text object.
        :return: The words of the text object."""
        return self._words

    @words.setter
    def words(self, string):
        """Set the words of the text object.
        :param string: The new words of the text object."""
        self._words = str(string)

    @property
    def font(self):
        """Get the font of the text object.
        :return: The font of the text object."""
        return self._font

    @font.setter
    def font(self, font_name):
        """Set the font of the text object.
        :param font_name: The new font of the text object."""
        self._font = str(font_name)

    @property
    def font_size(self):
        """Get the font size of the text object.
        :return: The font size of the text object."""
        return self._font_size

    @font_size.setter
    def font_size(self, size):
        self._font_size = size

    @property
    def color(self):
        """Get the color of the text object.
        :return: The color of the text object."""
        return self._color

    @color.setter
    def color(self, color_):
        self._color = color_
